/***************************************************************************
                          kvdr.cpp  -  description
                             -------------------
    begin                : Sat Aug  5 13:32:09 MEST 2000
    copyright            : (C) 2000 by Guido
    email                : gfiala@s.netic.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "kvdr.h"
#include <qmsgbox.h>
#include <qfile.h>
#include <kconfig.h>
#include <time.h>
#include "kvdr.xpm"

extern "C" {
//#define HAVE_BOOLEAN
#include <jpeglib.h>
}

// includes needed for X stuff
#include <X11/Xlib.h>
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86vmstr.h>
#include <X11/extensions/xf86vmode.h>

#ifdef USE_DPMS
#include <X11/extensions/dpms.h>
#endif

XF86VidModeModeInfo **vm_modelines;

//extract from dvb-videodev.h but can not include it here...
#define VIDEO_PALETTE_GREY		1	/* Linear greyscale */
#define VIDEO_PALETTE_HI240		2	/* High 240 cube (BT848) */
#define VIDEO_PALETTE_RGB565	3	/* 565 16 bit RGB */
#define VIDEO_PALETTE_RGB24		4	/* 24bit RGB */
#define VIDEO_PALETTE_RGB32		5	/* 32bit RGB */	
#define VIDEO_PALETTE_RGB555	6	/* 555 15bit RGB */
#define VIDEO_PALETTE_YUV422	7	/* YUV422 capture */
#define VIDEO_PALETTE_YUYV		8
#define VIDEO_PALETTE_UYVY		9	/* The great thing about standards is ... */
#define VIDEO_PALETTE_YUV420	10
#define VIDEO_PALETTE_YUV411	11	/* YUV411 capture */
#define VIDEO_PALETTE_RAW		  12	/* RAW capture (BT848) */
#define VIDEO_PALETTE_YUV422P	13	/* YUV 4:2:2 Planar */
#define VIDEO_PALETTE_YUV411P	14	/* YUV 4:1:1 Planar */
#define VIDEO_PALETTE_YUV420P	15	/* YUV 4:2:0 Planar */
#define VIDEO_PALETTE_YUV410P	16	/* YUV 4:1:0 Planar */
#define VIDEO_PALETTE_PLANAR	13	/* start of planar entries */
#define VIDEO_PALETTE_COMPONENT 7	/* start of component entries */

//XError-Handler:
XErrorHandler xerror(Display *, XErrorEvent *event)
{
  //should hide overlay now...
  return 0;
}

#define MIN_SIZE_X 160
#define MIN_SIZE_Y 120
#define MAX_X_PAL 768
#define MAX_Y_PAL 576
#define MAX_X_NTSC 640
#define MAX_Y_NTSC 480

//config-file-stuff:
#define GROUP_GCONFIG     "CONFIG__general_config"

Kvdr::Kvdr(QWidget *parent, const char *name)
: QWidget(parent, name,
  //should give us the required events, but crashs us instead: WType_Desktop
  Qt::WPaintClever | Qt::WResizeNoErase | Qt::WPaintUnclipped |
  Qt::WRepaintNoErase | Qt::WPaintDesktop )
{
  //cmdline-arguments:
  args = KCmdLineArgs::parsedArgs();
  //some Qt-Stuff:
  QColor bkgc(0,0,0);//Black is useful...
  setMinimumSize( MIN_SIZE_X, MIN_SIZE_Y );
  setBackgroundColor(bkgc);
  setIcon(QPixmap((const char **)kvdr));
  oldXErrorHandler= (void*)XSetErrorHandler((XErrorHandler) xerror);
  disp=qt_xdisplay();
  root=DefaultRootWindow(disp);
  screen=XDefaultScreen(disp);

  /* guess physical screen format */
  int width, bank, ram;
  base = NULL;
  if(!XF86DGAGetVideoLL(disp,screen,(int*)&base,&width,&bank,&ram))
  {
    base=NULL;
  }
  int n,i,pixmap_bytes=0;
  XPixmapFormatValues *pf = XListPixmapFormats(disp,&n);
  for (i = 0; i < n; i++) {
    if (pf[i].depth == QColor::numBitPlanes())
      pixmap_bytes = pf[i].bits_per_pixel/8;
  }
  //get color-depth from commandline:
  int tbpp=(args->getOption("b")).toUShort();
  if(tbpp==15 || tbpp==16 || tbpp==24 || tbpp==32)
  {
    bpp=tbpp;//yes, was given
  }
  else
  {
    bpp=pixmap_bytes*8;//as calculated
  }
  // little endian (intel) - dvb does'nt know about big endian anyway!?
  switch (pixmap_bytes)
  {
    case 1: palette = VIDEO_PALETTE_GREY; break;
    case 2: palette = (bpp==15) ? VIDEO_PALETTE_RGB555 :
                                  VIDEO_PALETTE_RGB565;
                                  break;
    case 3: palette = VIDEO_PALETTE_RGB24; break;
    case 4: palette = VIDEO_PALETTE_RGB32; break;
  }
  XWindowAttributes wts;
  if( XGetWindowAttributes(disp, root, &wts))
  {
    display_width=wts.width;display_height=wts.height;
  }
  else
  {
    display_width=0;display_height=0;
  }

  //get the available vidmodes:
  if(-1==XF86VidModeGetAllModeLines(disp,screen,&vm_count,&vm_modelines))
    perror("XF86VidModeGetAllModeLines");

  // now my status variables:
  ismute=false;  isfullscreen=false; ismultivideo=false;
  ovlstat=false; store=false; isPAL=true;dirty_position=true;
  ovlFbSet=false; ovlGeoSet=false; ovlStat=false;ovlClipCount=0;
  didRefresh=false;RefreshEntireScreen=false;cnfdlg=NULL;
  ShouldShowDialog=false;vdr_timeout=0;VidDev=0;vdr_mb_shown=false;
  brightness=40000;colour=33000;hue=33000;contrast=40000;
  video_aspect_ratio=4.0/3.0;
  aspect_ratio=video_aspect_ratio/
               (video_aspect_ratio-(float)display_width/display_height+1.0);

  i_size_x=200;i_size_y=(int)((float)i_size_x/aspect_ratio);
  i_g_size_x=isPAL?MAX_X_PAL:MAX_X_NTSC;i_g_size_y=isPAL?MAX_Y_PAL:MAX_Y_NTSC;
  i_jpg_qual=80;vdr_port=2001;Volume=0;FrontRear=50;MixerChannel=0;basename="";
  num_of_consecutive_expose_events=0;

  //Now overwrite it with .kvdrrc-values if available:
  load_kvdrrc();

  userKeys = new KAccel( this );

  //Todo: make this configurable out of .kvdrrc (via get-key or choice-list)
  userKeys->insertItem(i18n("Quit"),"Quit",Key_Q);
  userKeys->connectItem("Quit",this,SLOT(QuitSLOT()));

  userKeys->insertItem(i18n("Help"),"Help",Key_H);
  userKeys->connectItem("Help",this,SLOT(HelpSLOT()));

  userKeys->insertItem(i18n("Fullscreen"),"Fullscreen",Key_F);
  userKeys->connectItem("Fullscreen",this,SLOT(FullscreenSLOT()));

  userKeys->insertItem(i18n("Mute"),"Mute",Key_Enter);
  userKeys->connectItem("Mute",this,SLOT(MuteSLOT()));

  userKeys->insertItem(i18n("VolumeUp"),"VolumeUp",Key_Plus);
  userKeys->connectItem("VolumeUp",this,SLOT(VolumeUpSLOT()));

  userKeys->insertItem(i18n("VolumeDown"),"VolumeDown",Key_Minus);
  userKeys->connectItem("VolumeDown",this,SLOT(VolumeDownSLOT()));

  userKeys->insertItem(i18n("RearVolumeUp"),"RearVolumeUp",CTRL|Key_Plus);
  userKeys->connectItem("RearVolumeUp",this,SLOT(RearVolumeUpSLOT()));

  userKeys->insertItem(i18n("RearVolumeDown"),"RearVolumeDown",CTRL|Key_Minus);
  userKeys->connectItem("RearVolumeDown",this,SLOT(RearVolumeDownSLOT()));

  userKeys->insertItem(i18n("GrabJpg"),"GrabJpg",Key_J);
  userKeys->connectItem("GrabJpg",this,SLOT(GrabJpgSLOT()));

  userKeys->insertItem(i18n("GrabJpg"),"GrabJpg1",Key_Space);
  userKeys->connectItem("GrabJpg1",this,SLOT(GrabJpgSLOT()));

  userKeys->insertItem(i18n("GrabPnm"),"GrabPnm",Key_P);
  userKeys->connectItem("GrabPnm",this,SLOT(GrabPnmSLOT()));

  userKeys->insertItem(i18n("OverlayOnOff"),"OverlayOnOff",Key_O);
  userKeys->connectItem("OverlayOnOff",this,SLOT(OverlayOnOffSLOT(int)));

  userKeys->insertItem(i18n("Zap Video Dev"),"Zap Video Dev",Key_Z);
  userKeys->connectItem("Zap Video Dev",this,SLOT(ZapVideoDevSlot()));

  userKeys->insertItem(i18n("Multi Video Dev"),"Multi Video Dev",Key_X);
  userKeys->connectItem("Multi Video Dev",this,SLOT(MultiVideoDevSlot()));

  userKeys->insertItem(i18n("Configuration"),"Configuration",Key_C);
  userKeys->connectItem("Configuration",this,SLOT(ConfigurationSLOT()));

  connect(&AdjustSizeTimer,SIGNAL(timeout()),this,
          SLOT(AdjustAspectRatioSlot()));
  connect(&RefreshDisplayTimer,SIGNAL(timeout()),this,
          SLOT(RefreshDisplaySlot()));
  connect(&CheckClippingTimer,SIGNAL(timeout()),this,
          SLOT(CheckClippingTimerSlot()));
  //now the key-assignments for all the keys in vdr:
  userKeys->insertItem(i18n("MenuKey"),"MenuKey",Key_M);
  userKeys->connectItem("MenuKey",this,SLOT(KeyMenuSLOT()));
  userKeys->insertItem(i18n("OkKey"),"OkKey",Key_Return);
  userKeys->connectItem("OkKey",this,SLOT(KeyOkSLOT()));
  userKeys->insertItem(i18n("KeyLeft"),"KeyLeft",Key_Left);
  userKeys->connectItem("KeyLeft",this,SLOT(KeyLeftSLOT()));
  userKeys->insertItem(i18n("KeyRight"),"KeyRight",Key_Right);
  userKeys->connectItem("KeyRight",this,SLOT(KeyRightSLOT()));
  userKeys->insertItem(i18n("KeyUp"),"KeyUp",Key_Up);
  userKeys->connectItem("KeyUp",this,SLOT(KeyUpSLOT()));
  userKeys->insertItem(i18n("KeyDown"),"KeyDown",Key_Down);
  userKeys->connectItem("KeyDown",this,SLOT(KeyDownSLOT()));
  userKeys->insertItem(i18n("KeyBack"),"KeyBack",Key_Backspace);
  userKeys->connectItem("KeyBack",this,SLOT(KeyBackSLOT()));
  userKeys->insertItem(i18n("KeyRed"),"KeyRed",Key_F1);
  userKeys->connectItem("KeyRed",this,SLOT(KeyRedSLOT()));
  userKeys->insertItem(i18n("KeyGreen"),"KeyGreen",Key_F2);
  userKeys->connectItem("KeyGreen",this,SLOT(KeyGreenSLOT()));
  userKeys->insertItem(i18n("KeyYellow"),"KeyYellow",Key_F3);
  userKeys->connectItem("KeyYellow",this,SLOT(KeyYellowSLOT()));
  userKeys->insertItem(i18n("KeyBlue"),"KeyBlue",Key_F4);
  userKeys->connectItem("KeyBlue",this,SLOT(KeyBlueSLOT()));
  userKeys->insertItem(i18n("Key0"),"Key0",Key_0);
  userKeys->connectItem("Key0",this,SLOT(Key0SLOT()));
  userKeys->insertItem(i18n("Key1"),"Key1",Key_1);
  userKeys->connectItem("Key1",this,SLOT(Key1SLOT()));
  userKeys->insertItem(i18n("Key2"),"Key2",Key_2);
  userKeys->connectItem("Key2",this,SLOT(Key2SLOT()));
  userKeys->insertItem(i18n("Key3"),"Key3",Key_3);
  userKeys->connectItem("Key3",this,SLOT(Key3SLOT()));
  userKeys->insertItem(i18n("Key4"),"Key4",Key_4);
  userKeys->connectItem("Key4",this,SLOT(Key4SLOT()));
  userKeys->insertItem(i18n("Key5"),"Key5",Key_5);
  userKeys->connectItem("Key5",this,SLOT(Key5SLOT()));
  userKeys->insertItem(i18n("Key6"),"Key6",Key_6);
  userKeys->connectItem("Key6",this,SLOT(Key6SLOT()));
  userKeys->insertItem(i18n("Key7"),"Key7",Key_7);
  userKeys->connectItem("Key7",this,SLOT(Key7SLOT()));
  userKeys->insertItem(i18n("Key8"),"Key8",Key_8);
  userKeys->connectItem("Key8",this,SLOT(Key8SLOT()));
  userKeys->insertItem(i18n("Key9"),"Key9",Key_9);
  userKeys->connectItem("Key9",this,SLOT(Key9SLOT()));

  //Special-Functions ;-)
  userKeys->insertItem(i18n("Record"),"Record",Key_R);
  userKeys->connectItem("Record",this,SLOT(RecordSLOT()));
  userKeys->insertItem(i18n("Page Up"),"PgUp",Key_PageUp);
  userKeys->connectItem("PgUp",this,SLOT(PgUpSLOT()));
  userKeys->insertItem(i18n("Page Down"),"PgDown",Key_PageDown);
  userKeys->connectItem("PgDown",this,SLOT(PgDownSLOT()));

  //open the connection to vdr:
  svdrpc=new cSVDRPC();
  svdrpc->Open(vdr_port);

  //restore volume settings out of kvdrrc
  p << "amixer";
  connect(&p,SIGNAL( receivedStdout (KProcess *, char *, int )),this,
          SLOT( GotMixerGroups(KProcess *, char *, int ) ));
  p.start(KProcess::NotifyOnExit,KProcess::Stdout);
  StrMixerChannel="";

  //Disable Screensaver & APM, make this kvdrrc configurable...
#ifdef USE_DPMS
  DPMSInfo(disp, &dpms_state, &dpms_on);
#endif
  if(-1==XGetScreenSaver(disp,&ss_timeout,&ss_interval,&ss_prefer_blanking,&ss_allow_exposures))
    perror("XGetScreenSaver");

  f_size_x=isPAL?MAX_X_PAL:MAX_X_NTSC;f_size_y=(int)((float)f_size_x/aspect_ratio);
  if(f_size_y>(isPAL?MAX_Y_PAL:MAX_Y_NTSC))
  {
    f_size_y=isPAL?MAX_Y_PAL:MAX_Y_NTSC;f_size_x=(int)((float)f_size_y*aspect_ratio);
  }
  // set XEvents we want to receive
  const uint stdWidgetEventMask =                 // X event mask
        KeyPressMask | KeyReleaseMask |
        ButtonPressMask | ButtonReleaseMask |
        KeymapStateMask |
        ButtonMotionMask |
        EnterWindowMask | LeaveWindowMask |
        FocusChangeMask |
        ExposureMask | VisibilityChangeMask |
        StructureNotifyMask | SubstructureNotifyMask;

  // root window -- need visibility, sub-, structure
  if(-1==XSelectInput(disp,root, VisibilityChangeMask |
               StructureNotifyMask | SubstructureNotifyMask ))
    perror("XSelectInput-rootw");

  // tv window -- need visibility, structure
  if(-1==XSelectInput(disp,(Window) winId(), stdWidgetEventMask ))
    perror("XSelectInput-win");
	//call v4l-conf for each device once
  for(int i=0;i<=3;i++)
  {
  	KProcess v4l;
  	QString s;s.sprintf("%s%d",DEFAULT_VIDEO_DEV,i);
  	v4l << "v4l-conf" << "-c" << s;
    v4l.start(KProcess::Block,KProcess::Stderr);
  }
  //open video-device for overlay & grab capabilities:
  QString s; s.sprintf("%s%d",DEFAULT_VIDEO_DEV,VidDev);
  video_fd.setName((const char*)s);
  video_fd.open(IO_ReadWrite);
  if (video_fd.handle()==-1)
  {
	  QString s; s.sprintf("could not open %s%d\nCheck if other programs opened it\nthat you have rw-permission to the device\nor press 'C' and choose the right one.",
	                       DEFAULT_VIDEO_DEV,VidDev);
    QMessageBox::critical(this,"kvdr:",i18n(s));	
    //kapp->quit();
  }
  //Set the Frambuffer and Image properties:
  OvlF(display_width,display_height,(int)base,bpp,palette);
  OvlP(brightness,colour,hue,contrast);
  ovlstat=true;
  prevSize=geometry();
  setMaximumSize(f_size_x,f_size_y);
  resize(i_size_x,i_size_y);
  AdjustSizeTimer.start(10,true);//does correct Size now...
  CheckClippingTimer.start(100,false);
}

Kvdr::~Kvdr()
{
  //Workaround to wrong position on session restore:
  KConfig *config=kapp->config();
  config->setGroup(GROUP_GCONFIG);
  config->writeEntry("lp_x",pos().x());
  config->writeEntry("lp_y",pos().y());
  //undo anything:
  if(res_was_switched)
    XF86VidModeSwitchToMode(disp,screen,
                            vm_modelines[0]);
#ifdef USE_DPMS
  if(dpms_on) DPMSEnable(disp);//reenable only if previously enabled
#endif
  if(-1==XSetScreenSaver(disp,ss_timeout,ss_interval,ss_prefer_blanking,ss_allow_exposures))
    perror("XSetScreenSaver");
  XSetErrorHandler((XErrorHandler)oldXErrorHandler);
  OvlO(0);
  if(video_fd.handle()!=-1)
  {
  	video_fd.close();
  }
  svdrpc->CmdQuit();
  usleep(100000);//just wait to let vdr close the connection
  delete userKeys;
  delete svdrpc;
  if(cnfdlg) delete cnfdlg;
}

bool Kvdr::OvlF(int SizeX, int SizeY, int FbAddr, int Bpp, int Palette)
{
  // this is the problematic part!
  struct video_buffer vb;
  if(!ismultivideo)
  {
    int result = 0;
    if (video_fd.handle() < 0)
      return false;
    if (SizeX <= 0 || SizeY <= 0 || FbAddr == 0 || Bpp / 8 > 4 ||
        Bpp / 8 <= 0 || Palette <= 0 || Palette > 13 || ovlClipCount < 0 ||
        SizeX > 4096 || SizeY > 4096)
    {
       ovlFbSet = ovlGeoSet = false;
       OvlO(false);
       return false;
    }
    result |= ioctl(video_fd.handle(), VIDIOCGFBUF, &vb);
    if (result) perror("VIDIOCGFBUF");
//    vb.base = (void*)FbAddr;
//    vb.depth = Bpp;
//    vb.height = SizeY;
//    vb.width = SizeX;
//    vb.bytesperline = ((vb.depth + 1) / 8) * vb.width;
//    //now the real thing: setting the framebuffer
//    result |= ioctl(video_fd.handle(), VIDIOCSFBUF, &vb);
//    if (result) perror("VIDIOCSFBUF");
    if (vb.base!=(void*)FbAddr || vb.depth!=Bpp || vb.height!=SizeY ||
        vb.width!=SizeX || vb.bytesperline!=((vb.depth + 1) / 8) * vb.width
       )
    {
    	 fprintf(stderr,"kvdr and v4l-conf disagree about the framebuffer geometry\noverlay will not be shown\n");
       ovlFbSet = ovlGeoSet = false;
       ovlClipCount = 0;
       OvlO(false);
       return false;
    }
    else
    {
       ovlFbSizeX = SizeX;
       ovlFbSizeY = SizeY;
       ovlBpp = Bpp;
       ovlPalette = Palette;
       ovlFbSet = true;
       return true;
    }
  }
  else //ismultivideo
  {
//    for(int i=0;i<=3;i++)
//    {
//    	if(multivideo_fd[i].handle()!=-1)
//    	{
//        if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCGFBUF, &vb))
//        {
//        	QString s;s.sprintf("multi-VIDIOCGSFBUF %d",i);
//         	perror(s);
//        }
//        vb.base = (void*)FbAddr;
//        vb.depth = Bpp;
//        vb.height = SizeY;
//        vb.width = SizeX;
//        vb.bytesperline = ((vb.depth + 1) / 8) * vb.width;
        //now the real thing: setting the framebuffer
//        if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCSFBUF, &vb))
//        {
//        	QString s;s.sprintf("multi-VIDIOCSFBUF %d %x",i,(int)FbAddr);
//         	perror(s);
//        }
//      }
//    }
  }
  return true;
}

bool Kvdr::OvlP(int Brightness, int Colour, int Hue, int Contrast)
{
  struct video_picture vp;
  if(!ismultivideo)
  {
    if (video_fd.handle() < 0)
      return false;
    if (!ovlFbSet)
       return false;
    int result = 0;
    ovlBrightness = Brightness;
    ovlColour = Colour;
    ovlHue = Hue;
    ovlContrast = Contrast;
    result |= ioctl(video_fd.handle(), VIDIOCGPICT, &vp);
    vp.brightness = Brightness;
    vp.colour = Colour;
    vp.hue = Hue;
    vp.contrast = Contrast;
    vp.depth = ovlBpp;
    vp.palette = ovlPalette;
    result |= ioctl(video_fd.handle(), VIDIOCSPICT, &vp);
    return result == 0;
  }
  else //ismultivideo
  {
    for(int i=0;i<=3;i++)
    {
    	if(multivideo_fd[i].handle()!=-1)
    	{
        if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCGPICT, &vp))
        {
        	QString s;s.sprintf("multi-VIDIOCGPICT %d",i);
         	perror(s);
        }
        vp.brightness = Brightness;
        vp.colour = Colour;
        vp.hue = Hue;
        vp.contrast = Contrast;
        vp.depth = ovlBpp;
        vp.palette = ovlPalette;
        if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCSPICT, &vp))
        {
        	QString s;s.sprintf("multi-VIDIOCSPICT %d",i);
         	perror(s);
        }
      }
    }
  }
  return true;
}

bool Kvdr::OvlO(bool Value)
{
  int one = 1;
  int zero = 0;
	if (!ismultivideo)
	{
    if (video_fd.handle() < 0)
      return false;
    int result = 0;
    if (!ovlGeoSet && Value)
       return false;
    result |= ioctl(video_fd.handle(), VIDIOCCAPTURE, Value ? &one : &zero);
    if (result) perror("VIDIOCCAPTURE");
    ovlStat = Value;
    if (result)
    {
      ovlStat = false;
      return false;
    }
  }
  else //ismultivideo
  {
    for(int i=0;i<=3;i++)
    {
    	if(multivideo_fd[i].handle()!=-1)
    	{
        if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCCAPTURE, Value ? &one : &zero))
        {
        	QString s;s.sprintf("multi-VIDIOCCAPTURE %d",i);
         	perror(s);
        }
      }
    }
  }
  return true;
}

bool Kvdr::GImg(const char *FileName, bool Jpeg, int Quality, int SizeX, int SizeY)
{
  int result = 0;
  // just do this once?
  struct video_mbuf mbuf;
  for(int v=0;v<=3;v++)
  {
    if(ismultivideo?multivideo_fd[v].handle():video_fd.handle()!=-1)
    {
      result |= ioctl(ismultivideo?multivideo_fd[v].handle():video_fd.handle(), VIDIOCGMBUF, &mbuf);
      int msize = mbuf.size;
      unsigned char *mem = (unsigned char *)mmap(0, msize, PROT_READ | PROT_WRITE, MAP_SHARED, ismultivideo?multivideo_fd[v].handle():video_fd.handle(), 0);
      if (!mem || mem == (unsigned char *)-1)
         return false;
      // set up the size and RGB
      struct video_capability vc;
      result |= ioctl(ismultivideo?multivideo_fd[v].handle():video_fd.handle(), VIDIOCGCAP, &vc);
      struct video_mmap vm;
      vm.frame = 0;
      if ((SizeX > 0) && (SizeX <= vc.maxwidth) &&
          (SizeY > 0) && (SizeY <= vc.maxheight))
      {
        vm.width = SizeX;
        vm.height = SizeY;
      }
      else
      {
        vm.width = vc.maxwidth;
        vm.height = vc.maxheight;
      }
      vm.format = VIDEO_PALETTE_RGB24;
      // this needs to be done every time:
      result |= ioctl(ismultivideo?multivideo_fd[v].handle():video_fd.handle(), VIDIOCMCAPTURE, &vm);
      result |= ioctl(ismultivideo?multivideo_fd[v].handle():video_fd.handle(), VIDIOCSYNC, &vm.frame);
      // make RGB out of BGR:
      int memsize = vm.width * vm.height;
      unsigned char *mem1 = mem;
      for (int i = 0; i < memsize; i++)
      {
        unsigned char tmp = mem1[2];
        mem1[2] = mem1[0];
        mem1[0] = tmp;
        mem1 += 3;
      }
      if (Quality < 0)
         Quality = 255; //XXX is this 'best'???

      FILE *f;
      if(ismultivideo)
      {
      	//insert videodev for distinction:
        QString s,ext;
        s=FileName;
        ext=s.right(4);
        s=s.left(s.length()-4);
        s.sprintf("%s_Ovl%d%s",(const char*)s,v,(const char*)ext);
        f = fopen(s, "wb");
      }
      else
      {
        f = fopen(FileName, "wb");
      }
      if (f)
      {
        if (Jpeg)
        {
          // write JPEG file:
          struct jpeg_compress_struct cinfo;
          struct jpeg_error_mgr jerr;
          cinfo.err = jpeg_std_error(&jerr);
          jpeg_create_compress(&cinfo);
          jpeg_stdio_dest(&cinfo, f);
          cinfo.image_width = vm.width;
          cinfo.image_height = vm.height;
          cinfo.input_components = 3;
          cinfo.in_color_space = JCS_RGB;

          jpeg_set_defaults(&cinfo);
          jpeg_set_quality(&cinfo, Quality, true);
          jpeg_start_compress(&cinfo, true);

          int rs = vm.width * 3;
          JSAMPROW rp[vm.height];
          for (int k = 0; k < vm.height; k++)
              rp[k] = &mem[rs * k];
          jpeg_write_scanlines(&cinfo, rp, vm.height);
          jpeg_finish_compress(&cinfo);
          jpeg_destroy_compress(&cinfo);
        }
        else
        {
          // write PNM file:
          if (fprintf(f, "P6\n%d\n%d\n255\n", vm.width, vm.height) < 0 ||
              fwrite(mem, vm.width * vm.height * 3, 1, f) < 0)
          {
            result |= 1;
          }
        }
    	  fclose(f);
      }
      else
      {
        result |= 1;
      }
      munmap(mem, msize);
    }
    if(!ismultivideo)break;
  }
  if (ovlStat && ovlGeoSet)
  {
    // switch the Overlay on again (gf: why have i to do anything again?)
    OvlG(ovlSizeX, ovlSizeY, ovlPosX, ovlPosY);
  }
  if (ovlFbSet)
     OvlP(ovlBrightness, ovlColour, ovlHue, ovlContrast);

  return result == 0;
}

bool Kvdr::OvlG(int SizeX, int SizeY, int PosX, int PosY)
{
  struct video_capability vc;
	if (!ismultivideo)
	{
    if (video_fd.handle() < 0)
      return false;
    int result = 0;
    result |= ioctl(video_fd.handle(), VIDIOCGCAP, &vc);
    if (result) perror("VIDIOCGCAP");
    if (!ovlFbSet)
       return false;
    if (SizeX < vc.minwidth || SizeY < vc.minheight ||
        SizeX > vc.maxwidth || SizeY>vc.maxheight
       )
    {
      ovlGeoSet = false;
      OvlO(false);
      return false;
    }
    else
    {
      struct video_window vw;
      result |= ioctl(video_fd.handle(), VIDIOCGWIN,  &vw);
      if (result) perror("VIDIOCGWIN");
      vw.x = PosX;
      vw.y = PosY;
      vw.width = SizeX;
      vw.height = SizeY;
      vw.chromakey = ovlPalette;
      vw.flags = 0;
      vw.clips = ovlClipRects[0];
      vw.clipcount = ovlClipCount;
      result |= ioctl(video_fd.handle(), VIDIOCSWIN, &vw);
      if (result) perror("VIDIOCSWIN");
      if (result)
      {
        ovlGeoSet = false;
        ovlClipCount = 0;
        return false;
      }
      else
      {
         ovlSizeX = SizeX;
         ovlSizeY = SizeY;
         ovlPosX = PosX;
         ovlPosY = PosY;
         ovlGeoSet = true;
         ovlStat = true;
         return true;
      }
    }
  }
  else //ismultivideo!
  {
  	for(int i=0;i<=3;i++)
  	{
  		if(multivideo_fd[i].handle()!=-1)
  		{
        if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCGCAP, &vc))
        {
         	QString s;s.sprintf("multi-VIDIOCGCAP %d",i);
        	perror(s);
        }
        if (SizeX <= vc.maxwidth && SizeY<=vc.maxheight)
        {
      		int x=i&1?1:0;
      		int y=i&2?1:0;
          struct video_window vw;
          if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCGWIN,&vw))
          {
          	QString s;s.sprintf("multi-VIDIOCGWIN %d",i);
          	perror(s);
          }
          vw.x = PosX+x*SizeX/2;
          vw.y = PosY+y*SizeY/2;
          vw.width = SizeX/2;
          vw.height = SizeY/2;
          vw.chromakey = ovlPalette;
          vw.flags = 0;
          vw.clips = ovlClipRects[i];
          vw.clipcount = ovlClipCount;
          if(-1==ioctl(multivideo_fd[i].handle(), VIDIOCSWIN, &vw))
          {
          	QString s;s.sprintf("multi-VIDIOCSWIN %d %d %d %d %d %d %d %d\n",
          	        i,vw.x,vw.y,vw.width,vw.height,vw.chromakey,vw.flags,vw.clipcount);
          	perror(s);
          }
        }
      }
    }
    ovlSizeX = SizeX;
    ovlSizeY = SizeY;
    ovlPosX = PosX;
    ovlPosY = PosY;
    ovlGeoSet = true;
    ovlStat = true;
  }
  return true;
}

bool Kvdr::OvlC(int ClipCount, CRect *cr)
{
	if (!ismultivideo)
	{
    if (video_fd.handle() < 0)
      return false;
    if (ovlGeoSet && ovlFbSet)
    {
      for (int i = 0; i < ClipCount; i++)
      {
        ovlClipRects[i][0].x = cr[i].x;
        ovlClipRects[i][0].y = cr[i].y;
        ovlClipRects[i][0].width = cr[i].width;
        ovlClipRects[i][0].height = cr[i].height;
        ovlClipRects[i][0].next = &(ovlClipRects[i + 1][0]);
      }
      ovlClipCount = ClipCount;
      //use it:
      return OvlG(ovlSizeX, ovlSizeY, ovlPosX, ovlPosY);
    }
  }
  else //ismultivideo
  {
  	for(int i=0;i<=3;i++)
  	{
  		if(multivideo_fd[i].handle()!=-1)
  		{
        for (int j = 0; j < ClipCount; j++)
        {
        	int x=i&1?1:0;
        	int y=i&2?1:0;
          ovlClipRects[i][j].x = cr[j].x-ovlSizeX/2*x;
          ovlClipRects[i][j].y = cr[j].y-ovlSizeY/2*y;
          ovlClipRects[i][j].width = cr[j].width;
          ovlClipRects[i][j].height = cr[j].height;
          ovlClipRects[i][j].next = &(ovlClipRects[i][j + 1]);
        }
      }
    }
    ovlClipCount = ClipCount;
    OvlG(ovlSizeX, ovlSizeY, ovlPosX, ovlPosY);
  }
  return false;
}

void Kvdr::QuitSLOT()
{
  kapp->quit();
}

void Kvdr::ZapVideoDevSlot()
{
	if (ismultivideo) return;
  VidDev++;
  if (VidDev>3) VidDev=0;
	if (video_fd.handle()!=-1)
	{
		OverlayOnOffSLOT(-1);
		video_fd.close();
	}
	QString s;
	s.sprintf("%s%d",DEFAULT_VIDEO_DEV,VidDev);
	video_fd.setName((const char*)s);
	video_fd.open(IO_ReadWrite);
	if(-1!=video_fd.handle())
	{
    //now call some Functions to take them effect...
    OvlF(display_width,display_height,(int)base,bpp,palette);
    OvlP(brightness,colour,hue,contrast);
		OverlayOnOffSLOT(1);
  }
  else
  {
    //what?
    //QMessageBox::critical(this,"kvdr:","could not open device!");
  }
}

void Kvdr::MultiVideoDevSlot()
{
	ismultivideo=!ismultivideo;
	if(ismultivideo)
	{
		//open at maximum 4 videodev's
    video_fd.close();
		for(int i=0;i<=3;i++)
		{
			QString s;s.sprintf("%s%i",DEFAULT_VIDEO_DEV,i);
			multivideo_fd[i].setName(s);
			multivideo_fd[i].open(IO_ReadWrite);
		}
	}
	else
	{
		//close all but primary-videodev
		for(int i=0;i<=3;i++)
		{
			multivideo_fd[i].close();
		}
    video_fd.open(IO_ReadWrite);
	}
  OvlF(display_width,display_height,(int)base,bpp,palette);
	OvlP(brightness,colour,hue,contrast);
  OverlayOnOffSLOT(1);
}

void Kvdr::HelpSLOT()
{
  kapp->invokeHelp("","kvdr");
}

void Kvdr::FullscreenSLOT()
{
  isfullscreen=!isfullscreen;
  if(isfullscreen)
  {
    saveGeom=geometry();
    savePos=frameGeometry();
    // get the number of the fullscreen-modeline:
    int i;
    for (i = 0; i < vm_count; i++)
    {
   	  if ((isPAL?800:MAX_X_NTSC) == vm_modelines[i]->hdisplay &&
   	      (isPAL?600:MAX_Y_NTSC) == vm_modelines[i]->vdisplay)
   	   	  break;
    }
    if (i != 0 && i != vm_count)
    {
	    XF86VidModeSwitchToMode(disp,screen,
	    vm_modelines[i]);
	    res_was_switched= true;
    }
    else
    {
  	  res_was_switched=false;
    }
    XF86VidModeSetViewPort(disp,screen,0,0);
    OvlO(0);
    setMaximumSize(vm_modelines[i]->hdisplay,vm_modelines[i]->vdisplay);
    resize(vm_modelines[i]->hdisplay,vm_modelines[i]->vdisplay);
    setMinimumSize(vm_modelines[i]->hdisplay,vm_modelines[i]->vdisplay);
    move(savePos.x()-saveGeom.x(),savePos.y()-saveGeom.y());//failure in Qt!
    float aspect_ratio=video_aspect_ratio/1.33*
               ((float)vm_modelines[i]->hdisplay/
                        vm_modelines[i]->vdisplay);
    f_x=isPAL?MAX_X_PAL:MAX_X_NTSC,f_y=(int)((float)f_x/aspect_ratio);
    OvlG(f_x,f_y,(vm_modelines[i]->hdisplay-f_x)/2,
                 (vm_modelines[i]->vdisplay-f_y)/2);
    OvlO(1);ovlstat=true;//makes sense - doesn't it?
#ifdef USE_DPMS
    DPMSInfo(disp, &dpms_state, &dpms_on);
    DPMSDisable(disp);
#endif
    if(-1==XSetScreenSaver(disp,0,ss_interval,ss_prefer_blanking,ss_allow_exposures))
      perror("XSetScreenSaver-off");
    refreshDisplay(savePos);
    orgCursor=cursor();
    setCursor(blankCursor);
    orgpos=QCursor::pos();
    QCursor::setPos(0,0);
  }
  else
  {
    if(res_was_switched)
	     XF86VidModeSwitchToMode(disp,screen,
	                             vm_modelines[0]);
    OvlO(0);
    setMinimumSize( MIN_SIZE_X, MIN_SIZE_Y );
    move(savePos.x(),savePos.y());
    OvlG(saveGeom.width(),saveGeom.height(),saveGeom.x(),saveGeom.y());
    resize(saveGeom.width(),saveGeom.height());
    setMaximumSize(f_size_x,f_size_y);
   	QRect geom= geometry();//ohne Rahmen, sicher ist sicher!
    OvlG(geom.width(),geom.height(),geom.x(),geom.y());
	  OvlO(1);ovlstat=true;//not to bad?
#ifdef USE_DPMS
    if(dpms_on) DPMSEnable(disp);//reenable only if previously enabled
#endif
    if(-1==XSetScreenSaver(disp,ss_timeout,ss_interval,ss_prefer_blanking,ss_allow_exposures))
      perror("XSetScreenSaver-on");
    QWidget *d = kapp->desktop();
    QRect rect(0,0,d->width(),d->height());
    refreshDisplay(rect);
    setCursor(orgCursor);
    QCursor::setPos(orgpos);
    RefreshDisplayTimer.start(100,true);//does refresh display in 100ms
  }
}

void Kvdr::GetNextFilename(QString& fn, const char* basename, const char *ext)
{
  time_t now=time(NULL);
  localtime(&now);
  QTime temp;temp.start();
  QString CName;
  if (!ismultivideo) CName=svdrpc->CmdChan();
  else CName="";
  int n=0;
  QFile qf;
  do
  {
#ifdef USE_VFAT_FS
    fn.sprintf("%s%s_%d.%02d.%02d-%02d;%02d;%02d-%02d%s",basename,
#else
    fn.sprintf("%s%s_%d.%02d.%02d-%02d:%02d:%02d-%02d%s",basename,
#endif
                                     (const char*)CName,
                                     QDate::currentDate().year(),
                                     QDate::currentDate().month(),
                                     QDate::currentDate().day(),
                                     temp.hour(),
                                     temp.minute(),
                                     temp.second(),n,ext);
    qf.setName(fn);
  }while(qf.exists() && (n++<30));
}

void Kvdr::GrabJpgSLOT()
{
  QString fn;
  QString ext=".jpg";
  GetNextFilename(fn,basename,ext);
  GImg(fn,1,i_jpg_qual,i_g_size_x,i_g_size_y);
}

void Kvdr::GrabPnmSLOT()
{
  QString fn;
  QString ext=".pnm";
  GetNextFilename(fn,basename,ext);
  GImg(fn,0,0,i_g_size_x,i_g_size_y);
}

void Kvdr::OverlayOnOffSLOT(int state)
{
	switch(state)
	{
		case 0:  ovlstat=!ovlstat;break;
		case 1:  ovlstat=true;break;
		case -1: ovlstat=false;break;
		default: ovlstat=!ovlstat;break;
	}
  if(ovlstat)
  {
    QRect geom= geometry();//ohne Rahmen
    if(isfullscreen)
    {
      OvlG(f_x,f_y,(geom.width() -f_x)/2,
                              (geom.height()-f_y)/2);
    }
    else
    {
      OvlG(geom.width(),geom.height(),geom.x(),geom.y());
    }
    doClipping();
  }
  OvlO(ovlstat);
}

void Kvdr::ConfigurationSLOT()
{
    if(p.isRunning())
    {
      ShouldShowDialog=true;
      return;
    }
    if(!cnfdlg)cnfdlg= new ckvdrcfg(this,i18n("Kvdr configuration dialog"));
    cnfdlg->show();
    if(isfullscreen && res_was_switched)
    { //in fullscreen mode we need to move the cnf-dialog to the center of the viewport:
      cnfdlg->move(isPAL?(800-cnfdlg->maximumSize().width())/2:(640-cnfdlg->maximumSize().width())/2,
                   isPAL?(600-cnfdlg->maximumSize().height())/2:(480-cnfdlg->maximumSize().height())/2);
    }
    //somehow we lose this here, so we set it again:
    const uint stdWidgetEventMask =
        KeyPressMask | KeyReleaseMask |
        ButtonPressMask | ButtonReleaseMask |
        KeymapStateMask |
        ButtonMotionMask |
        FocusChangeMask |
        ExposureMask | VisibilityChangeMask |
        StructureNotifyMask | SubstructureNotifyMask|
        VisibilityChangeMask | StructureNotifyMask;
    // tv window -- need visibility, structure
    if(-1==XSelectInput(disp,(Window) winId(), stdWidgetEventMask ))
      perror("XSelectInput-winId");
    if(-1==XSelectInput(disp,root, VisibilityChangeMask |
              StructureNotifyMask | SubstructureNotifyMask ))
      perror("XSelectInput-winId");
}

void Kvdr::moveEvent(QMoveEvent * ev)
{
  //workaround for Session-Restore-Wrong-MoveEvent-Problem:
  if(dirty_position && (ev->pos().x()==0) && (ev->pos().y()==0))
  {
    move(lp_x,lp_y);
		dirty_position=false;
  }
  if(ovlstat)
  {
    QRect geom= geometry();//ohne Rahmen
    if(isfullscreen)
    {
      OvlG(f_x,f_y,(geom.width()-f_x)/2+geom.x(),
                   (geom.height()-f_y)/2+geom.y());
    }
    else
    {
      OvlG(geom.width(),geom.height(),geom.x(),geom.y());
    }
    refreshDisplay(prevSize);
    prevSize=geom;
    refreshDisplay(geom);
  }
}

void Kvdr::resizeEvent(QResizeEvent * ev)
{
  if(ovlstat && !isfullscreen)
  {
    QRect geom= geometry();//ohne Rahmen
    AdjustSizeTimer.start(100,true);//does correct Size in 100ms...
    OvlG(geom.width(),geom.height(),geom.x(),geom.y());
    refreshDisplay(prevSize);
    prevSize=geom;
    refreshDisplay(geom);
  }
}

void Kvdr::hideEvent(QHideEvent * ev)
{
  ovlstat=false;
}

void Kvdr::showEvent(QShowEvent * ev)
{
  ovlstat=true;
}

void Kvdr::AdjustAspectRatioSlot()
{
  if(ovlstat)
  {
    QRect geom= geometry();//ohne Rahmen
    float actARatio=(float)geom.width()/geom.height();
    if(!isfullscreen)
    {
      if( (actARatio>(aspect_ratio*1.01))
          || //let it have 1% failure without action...
          (actARatio<(aspect_ratio*0.99))
        )
      {
        //choose the closest shrinking-target-size:
        if(actARatio>aspect_ratio)
        { //it's a bit to long, make it shorter:
          geom.setWidth((int)((float)geom.height()*aspect_ratio));
        }
        else
        { //it's a bit to high, make it smaller:
          geom.setHeight((int)((float)geom.width()/aspect_ratio));
        }
        if(geometry()!=geom)
        { //only if we changed something:
		      resize(geom.width(),geom.height());
		      OvlG(geom.width(),geom.height(),geom.x(),geom.y());
		      refreshDisplay(prevSize);
  		    prevSize=geom;
	  	    refreshDisplay(geom);
        }
      }
    }
  }
}

#define safety 20
void Kvdr::refreshDisplay(const QRect& rect)
{
  XSetWindowAttributes xswa;
  unsigned long mask;
  Window rwin;

  xswa.override_redirect = true;
  xswa.backing_store = NotUseful;
  xswa.save_under = false;
  mask = ( CWBackingStore | CWSaveUnder |
           CWOverrideRedirect
         );

  rwin = XCreateWindow(disp, root, rect.x()-safety, rect.y()-safety,
                       rect.width()+2*safety,rect.height()+2*safety,
                       0, CopyFromParent, InputOutput, CopyFromParent,
                       mask, &xswa);
  XMapWindow(disp, rwin);
  XUnmapWindow(disp, rwin);
  XDestroyWindow(disp, rwin);
}

#define MAXSS 32767
#define MINSS (-32768)

void Kvdr::addCRect(int &cc, int x, int y, int w, int h)
{
  if( (x<=MAXSS) && (y<=MAXSS) && (w<=MAXSS) && (h<=MAXSS) &&
      (x>=MINSS) && (y>=MINSS) && (w>=MINSS) && (h>=MINSS) &&
      (cc<MAXCLIPRECTS)
    )
  {
    cr[cc].x=x;
    cr[cc].y=y;
    cr[cc].width=w;
    cr[cc].height=h;
    cc++;
  }
}

void Kvdr::CheckClippingTimerSlot()
{
  if(!svdrpc->Connected())
  {
      if(!vdr_mb_shown) //show only one message box ...
      {
        vdr_mb_shown=true;
  	    QMessageBox::critical(this,"kvdr:",i18n("could not connect to vdr-daemon,\n"\
                                       "check if vdr is running,\n"\
                                       "kvdr uses the same port (press 'C'),\n"\
                                       "that no other application like\n"\
                                       "telnet or another kvdr is connected\n"\
                                       "and if vdr is alive!"));
				svdrpc->Close();
				svdrpc->Open(vdr_port);
				vdr_mb_shown=false;
      }
  }
  //copy clipping-array for comparison
  memcpy(cr1,cr,sizeof(*cr)*MAXCLIPRECTS);
  bzero(cr,sizeof(*cr)*MAXCLIPRECTS);
  doClipping(0);
}

void Kvdr::doClipping(bool doit)
{
  if (!ovlstat) return;
  int wx, wy, ww, wh;
  int clipcount=0;
  int x,y,x2,y2;
  XWindowAttributes wts;
  Window parent, win, rroot,*children;
  uint nchildren, i;

  //take current geometry
  QRect geom=geometry();
  if(isfullscreen)
  {
    wx=geom.x()+(geom.width()-f_x)/2,
    wy=geom.y()+(geom.height()-f_y)/2;
    ww=geom.width();wh=geom.height();
  }
  else
  {
    wx=geom.x();wy=geom.y();ww=geom.width();wh=geom.height();
  }

  //hide parts of overlay that go beyond screen:
  if (wx<0)
  {
    addCRect(clipcount, 0, 0, (uint)(-wx), wh);
    RefreshEntireScreen=true;
  }
  if (wy<0)
  {
    addCRect(clipcount, 0, 0, ww, (uint)(-wy));
    RefreshEntireScreen=true;
  }
  if ((wx+ww) > display_width)
  {
    addCRect(clipcount, display_width-wx, 0, ww, wh);
    RefreshEntireScreen=true;
  }
  if ((wy+wh) > display_height)
  {
    addCRect(clipcount, 0, display_height-wy, ww, wh);
    RefreshEntireScreen=true;
  }

  win=(Window)winId();
  for (;;)
  {
    XQueryTree(disp, win, &rroot, &parent, &children, &nchildren);
    XFree((char *) children);
    if (root == parent) break;
    win = parent;
  }
  XQueryTree(disp, root, &rroot, &parent, &children, &nchildren);
  for (i = 0; i < nchildren; i++ )
    if ( children[i]==win ) break;
  for ( i++; i < nchildren; i++  )
  {
    XGetWindowAttributes(disp, children[i], &wts);
    if ( !(wts.map_state & IsViewable) ) continue;

    x=wts.x-wx;
    y=wts.y-wy;
    x2=x+wts.width+2*wts.border_width;
    y2=y+wts.height+2*wts.border_width;

    if ( (x2 >= 0) && (x < ww) && (y2 >= 0) && (y < wh) )
    {
      if ( x < 0 )   x=0;
      if ( y < 0 )   y=0;
      if ( x2 > ww ) x2=ww;
      if ( y2 > wh ) y2=wh;
      addCRect(clipcount, x, y, x2-x, y2-y );
    }
    else
      continue;
  }
  XFree((char *) children);
  if(0!=memcmp(cr1,cr,sizeof(*cr)*MAXCLIPRECTS) || doit)
  { //only clip new if something has changed or if forced!
	  OvlC(clipcount,cr);
//    if(!doit) RefreshEntireScreen=true;
    //if(!doit) refresh KWM::activeWindow () will be enough!
	  RefreshDisplayTimer.start(200,true);//does refresh display in 100ms
  	didRefresh=3;
  }
}

void Kvdr::RefreshDisplaySlot()
{
  if(RefreshEntireScreen)
  {
    QRect r(0,0,display_width,display_height);
    refreshDisplay(r);
    RefreshEntireScreen=false;
  }
  else
  {
    QRect r=geometry();
    refreshDisplay(r);
  }
}

//something for amixer:
void Kvdr::GotMixerGroups(KProcess *proc, char *buffer, int buflen)
{
  if(!p.isRunning() && p.normalExit() && proc && buffer && buflen)
  {
    //make QStrList out of "amixer" output:
    int i=0;
    for(;;)
    {
      //first find a Line starting with "Group"
      char *n=strstr(buffer+i,"Group");
      if(n==NULL)break;//nothing more!
      char *nn=strstr(n,",");
      if(nn>0)
      {
        i=nn-buffer;//start up next time
      }
      else
      {
        break;//nothing more!
      }
      if((n>0)&&(nn>0))
      { //found Group-Line
        QString t(n+7);t.truncate(nn-n-8);
        mixerdevlist.append(t);
      }
      else break;
    }
    //hopefully the number has'nt changed:
    if(mixerdevlist.count()>=MixerChannel)
    {
      StrMixerChannel=mixerdevlist.at(MixerChannel);
      ismute=true;MuteSLOT();//set-unmute!
      SetVolume(Volume,StrMixerChannel,FrontRear);//restore Volume
    }
  }
  if(ShouldShowDialog) ConfigurationSLOT();
}

void Kvdr::MuteSLOT()
{
  ismute=!ismute;
  if (!mixerp.isRunning())
  {
    mixerp.clearArguments();
    mixerp.setExecutable("amixer");
    if(ismute)
      mixerp << "set" << StrMixerChannel << "mute";
    else
      mixerp << "set" << StrMixerChannel << "unmute";
    mixerp.start(KProcess::DontCare,KProcess::Stdout);
    mixerp1.clearArguments();
    mixerp1.setExecutable("amixer");
    if(ismute)
      mixerp1 << "set" << "'Rear'" << "mute";
    else
      mixerp1 << "set" << "'Rear'" << "unmute";
    mixerp1.start(KProcess::DontCare,KProcess::Stdout);
  }
}

void Kvdr::VolumeUpSLOT()
{
  if (!mixerp.isRunning())
  {
    if(Volume<100)Volume++;
		SetVolume(Volume,StrMixerChannel,FrontRear);
  }
}

void Kvdr::VolumeDownSLOT()
{
  if (!mixerp.isRunning())
  {
    if(Volume>0)Volume--;
		SetVolume(Volume,StrMixerChannel,FrontRear);
  }
}

void Kvdr::RearVolumeUpSLOT()
{
  if (!mixerp.isRunning())
  {
    if(FrontRear<100)FrontRear++;
		SetVolume(Volume,StrMixerChannel,FrontRear);
  }
}

void Kvdr::RearVolumeDownSLOT()
{
  if (!mixerp.isRunning())
  {
    if(FrontRear>0)FrontRear--;
		SetVolume(Volume,StrMixerChannel,FrontRear);
  }
}

void Kvdr::SetVolume(int Volume, QString MGroup, int FrontRear)
{
  if (!mixerp2.isRunning() && (Volume<=100) && (Volume>=0))
  {
    if(ismute)MuteSLOT();
		int FrontVolume,RearVolume;
    FrontVolume=(int)((double)(100-FrontRear)*0.02*Volume);
    RearVolume =(int)((double)(FrontRear    )*0.02*Volume);
    if(FrontVolume>100)FrontVolume=100;
    if(RearVolume >100)RearVolume =100;
    QString t;t.sprintf("%d%%",FrontVolume);
    QString t1;t1.sprintf("%d%%",RearVolume);
    mixerp2.clearArguments();
    mixerp2.setExecutable("amixer");
    mixerp2 << "set" << MGroup << t << t;
    mixerp2.start(KProcess::DontCare,KProcess::Stdout);
    mixerp3.clearArguments();
    mixerp3.setExecutable("amixer");
    mixerp3 << "set" << "'Rear'" << t1 << t1;
    mixerp3.start(KProcess::DontCare,KProcess::Stdout);
  }
}

void Kvdr::load_kvdrrc()
{
  KConfig *config=kapp->config();

  config->setGroup(GROUP_GCONFIG);

  isPAL=config->readNumEntry("isPal",isPAL);
  video_aspect_ratio=config->readDoubleNumEntry("video_ascpect_ratio",
                                                video_aspect_ratio);
  lp_x=config->readNumEntry("lp_x",lp_x);
  lp_y=config->readNumEntry("lp_y",lp_y);
  i_size_x=config->readNumEntry("i_size_x",i_size_x);
  i_size_y=config->readNumEntry("i_size_y",i_size_y);
  i_g_size_x=config->readNumEntry("i_g_size_x",i_g_size_x);
  i_g_size_y=config->readNumEntry("i_g_size_y",i_g_size_y);
  i_jpg_qual=config->readNumEntry("i_jpg_qual",i_jpg_qual);
  vdr_port=config->readNumEntry("vdr_port",vdr_port);
  colour=config->readNumEntry("colour",colour);
  brightness=config->readNumEntry("brightness",brightness);
  hue=config->readNumEntry("hue",hue);
  contrast=config->readNumEntry("contrast",contrast);
  basename=config->readEntry("basename",basename);
  Volume=config->readNumEntry("Volume",Volume);
  FrontRear=config->readNumEntry("FrontRear",FrontRear);
  MixerChannel=config->readNumEntry("MixerChannel",MixerChannel);
  VidDev=config->readNumEntry("VidDev",VidDev);
  if(config->readNumEntry("ShowHelp",1))
  { //automatically invoke HTML-help the very first time:
    kapp->invokeHelp("index-3.html","kvdr");
    config->writeEntry("ShowHelp",0);
  }
}

void Kvdr::save_kvdrrc()
{
  KConfig *config=kapp->config();

  config->setGroup(GROUP_GCONFIG);

  config->writeEntry("isPal",isPAL);
  config->writeEntry("video_ascpect_ratio",video_aspect_ratio);
  config->writeEntry("i_size_x",i_size_x);
  config->writeEntry("i_size_y",i_size_y);
  config->writeEntry("i_g_size_x",i_g_size_x);
  config->writeEntry("i_g_size_y",i_g_size_y);
  config->writeEntry("i_jpg_qual",i_jpg_qual);
  config->writeEntry("vdr_port",vdr_port);
  config->writeEntry("colour",colour);
  config->writeEntry("brightness",brightness);
  config->writeEntry("hue",hue);
  config->writeEntry("contrast",contrast);
  config->writeEntry("basename",basename);
  config->writeEntry("Volume",Volume);
  config->writeEntry("FrontRear",FrontRear);
  config->writeEntry("MixerChannel",MixerChannel);
  config->writeEntry("VidDev",VidDev);
}

//now the tricky X11-Event-stuff:

bool Kvdr::x11Event( XEvent *event )
{
  if ( isVisible() && ovlstat && (event->xvisibility.window == (Window) winId()) )
  {
    if(event->type==Expose)
    {num_of_consecutive_expose_events++;}
    else
    {num_of_consecutive_expose_events=0;}
    switch ( event->type )
    {
    case Expose:
      {
          if(didRefresh>0)
          {
						//nothing to do...
						//printf("Expose>0\n");fflush(stdout);
            //did something moved in front of kvdr?
            if(num_of_consecutive_expose_events>4)
            {
              RefreshEntireScreen=true;
            }
          }
					else if ( !event->xexpose.count )
					{
					  //printf("Exposed\n");fflush(stdout);
            doClipping();
					}
      }
      break;
    case VisibilityNotify:
      newVisibility= event->xvisibility.state;
      switch (event->xvisibility.state)
      {
        case VisibilityFullyObscured:
        case VisibilityPartiallyObscured:
        case VisibilityUnobscured:
             if(didRefresh>0){didRefresh--;break;}
             doClipping();
             //printf("Visibility %d\n",event->type);fflush(stdout);
	           break;
      }
      break;
    case UnmapNotify:
      OvlO(0);
      RefreshDisplayTimer.start(10,true);//does refresh display in 10ms
      break;
    case MapNotify:
      OvlO(1);
      break;
		case 10:
		case 33:
						 doClipping();
	           break;
    //default:fprintf(stderr,"no-event %d\n",event->type);
    }
  }
  return false;
}
